@@ -44,6 +44,7 @@ static void backend_destroy(struct wlr_backend *backend) {
 	wlr_signal_emit_safe(&backend->events.destroy, backend);
 
 	wl_list_remove(&drm->display_destroy.link);
+	wl_list_remove(&drm->session_destroy.link);
 	wl_list_remove(&drm->session_signal.link);
 	wl_list_remove(&drm->drm_invalidated.link);
 
@@ -92,7 +93,7 @@ static void session_signal(struct wl_listener *listener, void *data) {
 
 		struct wlr_drm_connector *conn;
 		wl_list_for_each(conn, &drm->outputs, link){
-			if (conn->output.enabled) {
+			if (conn->output.enabled && conn->output.current_mode != NULL) {
 				drm_connector_set_mode(&conn->output,
 						conn->output.current_mode);
 			} else {
@@ -135,6 +136,12 @@ static void drm_invalidated(struct wl_listener *listener, void *data) {
 	scan_drm_connectors(drm);
 }
 
+static void handle_session_destroy(struct wl_listener *listener, void *data) {
+	struct wlr_drm_backend *drm =
+		wl_container_of(listener, drm, session_destroy);
+	backend_destroy(&drm->backend);
+}
+
 static void handle_display_destroy(struct wl_listener *listener, void *data) {
 	struct wlr_drm_backend *drm =
 		wl_container_of(listener, drm, display_destroy);
@@ -197,6 +204,9 @@ struct wlr_backend *wlr_drm_backend_create(struct wl_display *display,
 		goto error_event;
 	}
 
+	drm->session_destroy.notify = handle_session_destroy;
+	wl_signal_add(&session->events.destroy, &drm->session_destroy);
+
 	drm->display_destroy.notify = handle_display_destroy;
 	wl_display_add_destroy_listener(display, &drm->display_destroy);
 
