@@ -1,4 +1,5 @@
 #include <assert.h>
+#include <drm_fourcc.h>
 #include <gbm.h>
 #include <stdbool.h>
 #include <stdlib.h>
@@ -10,7 +11,6 @@
 #include <wlr/types/wlr_matrix.h>
 #include <wlr/util/log.h>
 #include "backend/drm/drm.h"
-#include "glapi.h"
 
 bool init_drm_renderer(struct wlr_drm_backend *drm,
 		struct wlr_drm_renderer *renderer, wlr_renderer_create_func_t create_renderer_func) {
@@ -61,7 +61,7 @@ void finish_drm_renderer(struct wlr_drm_renderer *renderer) {
 
 bool init_drm_surface(struct wlr_drm_surface *surf,
 		struct wlr_drm_renderer *renderer, uint32_t width, uint32_t height,
-		uint32_t format, uint32_t flags) {
+		uint32_t format, struct wlr_drm_format_set *set, uint32_t flags) {
 	if (surf->width == width && surf->height == height) {
 		return true;
 	}
@@ -80,11 +80,23 @@ bool init_drm_surface(struct wlr_drm_surface *surf,
 			surf->back = NULL;
 		}
 		gbm_surface_destroy(surf->gbm);
+		surf->gbm = NULL;
 	}
 	wlr_egl_destroy_surface(&surf->renderer->egl, surf->egl);
 
-	surf->gbm = gbm_surface_create(renderer->gbm, width, height,
-		format, GBM_BO_USE_RENDERING | flags);
+	if (!(flags & GBM_BO_USE_LINEAR) && set != NULL) {
+		const struct wlr_drm_format *drm_format =
+			wlr_drm_format_set_get(set, format);
+		if (drm_format != NULL) {
+			surf->gbm = gbm_surface_create_with_modifiers(renderer->gbm,
+				width, height, format, drm_format->modifiers, drm_format->len);
+		}
+	}
+
+	if (surf->gbm == NULL) {
+		surf->gbm = gbm_surface_create(renderer->gbm, width, height,
+			format, GBM_BO_USE_RENDERING | flags);
+	}
 	if (!surf->gbm) {
 		wlr_log_errno(WLR_ERROR, "Failed to create GBM surface");
 		goto error_zero;
@@ -165,26 +177,40 @@ void post_drm_surface(struct wlr_drm_surface *surf) {
 
 struct gbm_bo *import_gbm_bo(struct wlr_drm_renderer *renderer,
 		struct wlr_dmabuf_attributes *attribs) {
-	struct gbm_import_fd_modifier_data data = {
-		.width = attribs->width,
-		.height = attribs->height,
-		.format = attribs->format,
-		.num_fds = attribs->n_planes,
-		.modifier = attribs->modifier,
-	};
+	if (attribs->modifier == DRM_FORMAT_MOD_INVALID && attribs->n_planes == 1
+			&& attribs->offset[0] == 0) {
+		struct gbm_import_fd_data data = {
+			.fd = attribs->fd[0],
+			.width = attribs->width,
+			.height = attribs->height,
+			.stride = attribs->stride[0],
+			.format = attribs->format,
+		};
 
-	if ((size_t)attribs->n_planes > sizeof(data.fds) / sizeof(data.fds[0])) {
-		return NULL;
+		return gbm_bo_import(renderer->gbm, GBM_BO_IMPORT_FD,
+			&data, GBM_BO_USE_SCANOUT);
+	} else {
+		struct gbm_import_fd_modifier_data data = {
+			.width = attribs->width,
+			.height = attribs->height,
+			.format = attribs->format,
+			.num_fds = attribs->n_planes,
+			.modifier = attribs->modifier,
+		};
+
+		if ((size_t)attribs->n_planes > sizeof(data.fds) / sizeof(data.fds[0])) {
+			return NULL;
+		}
+
+		for (size_t i = 0; i < (size_t)attribs->n_planes; ++i) {
+			data.fds[i] = attribs->fd[i];
+			data.strides[i] = attribs->stride[i];
+			data.offsets[i] = attribs->offset[i];
+		}
+
+		return gbm_bo_import(renderer->gbm, GBM_BO_IMPORT_FD_MODIFIER,
+			&data, GBM_BO_USE_SCANOUT);
 	}
-
-	for (size_t i = 0; i < (size_t)attribs->n_planes; ++i) {
-		data.fds[i] = attribs->fd[i];
-		data.strides[i] = attribs->stride[i];
-		data.offsets[i] = attribs->offset[i];
-	}
-
-	return gbm_bo_import(renderer->gbm, GBM_BO_IMPORT_FD_MODIFIER,
-		&data, GBM_BO_USE_SCANOUT);
 }
 
 bool export_drm_bo(struct gbm_bo *bo, struct wlr_dmabuf_attributes *attribs) {
@@ -261,19 +287,22 @@ struct gbm_bo *copy_drm_surface_mgpu(struct wlr_drm_surface *dest,
 
 bool init_drm_plane_surfaces(struct wlr_drm_plane *plane,
 		struct wlr_drm_backend *drm, int32_t width, uint32_t height,
-		uint32_t format) {
+		uint32_t format, bool with_modifiers) {
+	struct wlr_drm_format_set *format_set =
+		with_modifiers ? &plane->formats : NULL;
+
 	if (!drm->parent) {
 		return init_drm_surface(&plane->surf, &drm->renderer, width, height,
-			format, GBM_BO_USE_SCANOUT);
+			format, format_set, GBM_BO_USE_SCANOUT);
 	}
 
 	if (!init_drm_surface(&plane->surf, &drm->parent->renderer,
-			width, height, format, GBM_BO_USE_LINEAR)) {
+			width, height, format, NULL, GBM_BO_USE_LINEAR)) {
 		return false;
 	}
 
 	if (!init_drm_surface(&plane->mgpu_surf, &drm->renderer,
-			width, height, format, GBM_BO_USE_SCANOUT)) {
+			width, height, format, format_set, GBM_BO_USE_SCANOUT)) {
 		finish_drm_surface(&plane->surf);
 		return false;
 	}
