@@ -42,7 +42,7 @@ static void handle_tablet_tool_v2_set_cursor(struct wl_client *client,
 		.hotspot_x = hotspot_x,
 		.hotspot_y = hotspot_y,
 		.seat_client = tool->seat->seat_client,
-		};
+	};
 
 	wl_signal_emit(&tool->tool->events.set_cursor, &evt);
 }
@@ -73,6 +73,9 @@ static enum zwp_tablet_tool_v2_type tablet_type_from_wlr_type(
 		return ZWP_TABLET_TOOL_V2_TYPE_MOUSE;
 	case WLR_TABLET_TOOL_TYPE_LENS:
 		return ZWP_TABLET_TOOL_V2_TYPE_LENS;
+	default:
+		/* We skip these devices earlier on */
+		assert(false && "Unreachable");
 	}
 
 	assert(false && "Unreachable");
@@ -196,6 +199,21 @@ struct wlr_tablet_v2_tablet_tool *wlr_tablet_tool_create(
 		struct wlr_tablet_manager_v2 *manager,
 		struct wlr_seat *wlr_seat,
 		struct wlr_tablet_tool *wlr_tool) {
+	switch (wlr_tool->type) {
+	case WLR_TABLET_TOOL_TYPE_PEN:
+	case WLR_TABLET_TOOL_TYPE_ERASER:
+	case WLR_TABLET_TOOL_TYPE_BRUSH:
+	case WLR_TABLET_TOOL_TYPE_PENCIL:
+	case WLR_TABLET_TOOL_TYPE_AIRBRUSH:
+	case WLR_TABLET_TOOL_TYPE_MOUSE:
+	case WLR_TABLET_TOOL_TYPE_LENS:
+		/* supported */
+		break;
+	default:
+		/* Unsupported */
+		return NULL;
+	}
+
 	struct wlr_tablet_seat_v2 *seat = get_or_create_tablet_seat(manager, wlr_seat);
 	if (!seat) {
 		return NULL;
@@ -350,7 +368,7 @@ void wlr_send_tablet_v2_tablet_tool_proximity_in(
 
 	tool->current_client = tool_client;
 
-	uint32_t serial = wl_display_next_serial(wl_client_get_display(client));
+	uint32_t serial = wlr_seat_client_next_serial(tool_client->seat->seat_client);
 	tool->focused_surface = surface;
 	tool->proximity_serial = serial;
 
@@ -398,6 +416,7 @@ void wlr_send_tablet_v2_tablet_tool_proximity_out(
 			send_tool_frame(tool->current_client);
 		}
 		zwp_tablet_tool_v2_send_proximity_out(tool->current_client->resource);
+		send_tool_frame(tool->current_client);
 
 		tool->current_client = NULL;
 		tool->focused_surface = NULL;
@@ -466,9 +485,8 @@ void wlr_send_tablet_v2_tablet_tool_button(
 	ssize_t index = tablet_tool_button_update(tool, button, state);
 
 	if (tool->current_client) {
-		struct wl_client *client =
-			wl_resource_get_client(tool->current_client->resource);
-		uint32_t serial = wl_display_next_serial(wl_client_get_display(client));
+		uint32_t serial = wlr_seat_client_next_serial(
+			tool->current_client->seat->seat_client);
 		if (index >= 0) {
 			tool->pressed_serials[index] = serial;
 		}
@@ -496,9 +514,8 @@ void wlr_send_tablet_v2_tablet_tool_down(struct wlr_tablet_v2_tablet_tool *tool)
 
 	tool->is_down = true;
 	if (tool->current_client) {
-		struct wl_client *client =
-			wl_resource_get_client(tool->current_client->resource);
-		uint32_t serial = wl_display_next_serial(wl_client_get_display(client));
+		uint32_t serial = wlr_seat_client_next_serial(
+			tool->current_client->seat->seat_client);
 
 		zwp_tablet_tool_v2_send_down(tool->current_client->resource,
 			serial);
