@@ -165,10 +165,6 @@ static const struct zwlr_export_dmabuf_manager_v1_interface manager_impl = {
 	.destroy = manager_handle_destroy,
 };
 
-static void manager_handle_resource_destroy(struct wl_resource *resource) {
-	wl_list_remove(wl_resource_get_link(resource));
-}
-
 static void manager_bind(struct wl_client *client, void *data, uint32_t version,
 		uint32_t id) {
 	struct wlr_export_dmabuf_manager_v1 *manager = data;
@@ -180,15 +176,16 @@ static void manager_bind(struct wl_client *client, void *data, uint32_t version,
 		return;
 	}
 	wl_resource_set_implementation(resource, &manager_impl, manager,
-		manager_handle_resource_destroy);
-
-	wl_list_insert(&manager->resources, wl_resource_get_link(resource));
+		NULL);
 }
 
 static void handle_display_destroy(struct wl_listener *listener, void *data) {
 	struct wlr_export_dmabuf_manager_v1 *manager =
 		wl_container_of(listener, manager, display_destroy);
-	wlr_export_dmabuf_manager_v1_destroy(manager);
+	wlr_signal_emit_safe(&manager->events.destroy, manager);
+	wl_list_remove(&manager->display_destroy.link);
+	wl_global_destroy(manager->global);
+	free(manager);
 }
 
 struct wlr_export_dmabuf_manager_v1 *wlr_export_dmabuf_manager_v1_create(
@@ -198,7 +195,6 @@ struct wlr_export_dmabuf_manager_v1 *wlr_export_dmabuf_manager_v1_create(
 	if (manager == NULL) {
 		return NULL;
 	}
-	wl_list_init(&manager->resources);
 	wl_list_init(&manager->frames);
 	wl_signal_init(&manager->events.destroy);
 
@@ -215,22 +211,3 @@ struct wlr_export_dmabuf_manager_v1 *wlr_export_dmabuf_manager_v1_create(
 
 	return manager;
 }
-
-void wlr_export_dmabuf_manager_v1_destroy(
-		struct wlr_export_dmabuf_manager_v1 *manager) {
-	if (manager == NULL) {
-		return;
-	}
-	wlr_signal_emit_safe(&manager->events.destroy, manager);
-	wl_list_remove(&manager->display_destroy.link);
-	wl_global_destroy(manager->global);
-	struct wl_resource *resource, *resource_tmp;
-	wl_resource_for_each_safe(resource, resource_tmp, &manager->resources) {
-		wl_resource_destroy(resource);
-	}
-	struct wlr_export_dmabuf_frame_v1 *frame, *frame_tmp;
-	wl_list_for_each_safe(frame, frame_tmp, &manager->frames, link) {
-		wl_resource_destroy(frame->resource);
-	}
-	free(manager);
-}
