@@ -10,6 +10,7 @@
 
 #define OUTPUT_MANAGER_VERSION 3
 #define OUTPUT_DONE_DEPRECATED_SINCE_VERSION 3
+#define OUTPUT_DESCRIPTION_MUTABLE_SINCE_VERSION 3
 
 static void output_handle_destroy(struct wl_client *client,
 		struct wl_resource *resource) {
@@ -70,6 +71,7 @@ static void output_destroy(struct wlr_xdg_output_v1 *output) {
 		wl_list_init(wl_resource_get_link(resource));
 	}
 	wl_list_remove(&output->destroy.link);
+	wl_list_remove(&output->description.link);
 	wl_list_remove(&output->link);
 	free(output);
 }
@@ -124,14 +126,15 @@ static void output_manager_handle_get_xdg_output(struct wl_client *client,
 	if (version >= ZXDG_OUTPUT_V1_NAME_SINCE_VERSION) {
 		zxdg_output_v1_send_name(xdg_output_resource, output->name);
 	}
-	if (version >= ZXDG_OUTPUT_V1_DESCRIPTION_SINCE_VERSION) {
-		char description[128];
-		snprintf(description, sizeof(description), "%s %s %s (%s)",
-			output->make, output->model, output->serial, output->name);
-		zxdg_output_v1_send_description(xdg_output_resource, description);
+	if (version >= ZXDG_OUTPUT_V1_DESCRIPTION_SINCE_VERSION &&
+			output->description != NULL) {
+		zxdg_output_v1_send_description(xdg_output_resource,
+			output->description);
 	}
 
 	output_send_details(xdg_output, xdg_output_resource);
+
+	wl_output_send_done(output_resource);
 }
 
 static const struct zxdg_output_manager_v1_interface
@@ -140,11 +143,6 @@ static const struct zxdg_output_manager_v1_interface
 	.get_xdg_output = output_manager_handle_get_xdg_output,
 };
 
-static void output_manager_handle_resource_destroy(
-		struct wl_resource *resource) {
-	wl_list_remove(wl_resource_get_link(resource));
-}
-
 static void output_manager_bind(struct wl_client *wl_client, void *data,
 		uint32_t version, uint32_t id) {
 	struct wlr_xdg_output_manager_v1 *manager = data;
@@ -156,8 +154,7 @@ static void output_manager_bind(struct wl_client *wl_client, void *data,
 		return;
 	}
 	wl_resource_set_implementation(resource, &output_manager_implementation,
-		manager, output_manager_handle_resource_destroy);
-	wl_list_insert(&manager->resources, wl_resource_get_link(resource));
+		manager, NULL);
 }
 
 static void handle_output_destroy(struct wl_listener *listener, void *data) {
@@ -165,6 +162,25 @@ static void handle_output_destroy(struct wl_listener *listener, void *data) {
 	output_destroy(output);
 }
 
+static void handle_output_description(struct wl_listener *listener,
+		void *data) {
+	struct wlr_xdg_output_v1 *xdg_output =
+		wl_container_of(listener, xdg_output, description);
+	struct wlr_output *output = xdg_output->layout_output->output;
+
+	if (output->description == NULL) {
+		return;
+	}
+
+	struct wl_resource *resource;
+	wl_resource_for_each(resource, &xdg_output->resources) {
+		if (wl_resource_get_version(resource) >=
+				OUTPUT_DESCRIPTION_MUTABLE_SINCE_VERSION) {
+			zxdg_output_v1_send_description(resource, output->description);
+		}
+	}
+}
+
 static void add_output(struct wlr_xdg_output_manager_v1 *manager,
 		struct wlr_output_layout_output *layout_output) {
 	struct wlr_xdg_output_v1 *output = calloc(1, sizeof(struct wlr_xdg_output_v1));
@@ -176,6 +192,9 @@ static void add_output(struct wlr_xdg_output_manager_v1 *manager,
 	output->layout_output = layout_output;
 	output->destroy.notify = handle_output_destroy;
 	wl_signal_add(&layout_output->events.destroy, &output->destroy);
+	output->description.notify = handle_output_description;
+	wl_signal_add(&layout_output->output->events.description,
+		&output->description);
 	wl_list_insert(&manager->outputs, &output->link);
 	output_update(output);
 }
@@ -201,16 +220,29 @@ static void handle_layout_change(struct wl_listener *listener, void *data) {
 	output_manager_send_details(manager);
 }
 
+static void manager_destroy(struct wlr_xdg_output_manager_v1 *manager) {
+	wlr_signal_emit_safe(&manager->events.destroy, manager);
+	wl_list_remove(&manager->display_destroy.link);
+	wl_list_remove(&manager->layout_add.link);
+	wl_list_remove(&manager->layout_change.link);
+	wl_list_remove(&manager->layout_destroy.link);
+	free(manager);
+}
+
 static void handle_layout_destroy(struct wl_listener *listener, void *data) {
 	struct wlr_xdg_output_manager_v1 *manager =
 		wl_container_of(listener, manager, layout_destroy);
-	wlr_xdg_output_manager_v1_destroy(manager);
+	manager_destroy(manager);
+}
+
+static void handle_display_destroy(struct wl_listener *listener, void *data) {
+	struct wlr_xdg_output_manager_v1 *manager =
+		wl_container_of(listener, manager, display_destroy);
+	manager_destroy(manager);
 }
 
 struct wlr_xdg_output_manager_v1 *wlr_xdg_output_manager_v1_create(
 		struct wl_display *display, struct wlr_output_layout *layout) {
-	assert(display && layout);
-
 	// TODO: require wayland-protocols 1.18 and remove this condition
 	int version = OUTPUT_MANAGER_VERSION;
 	if (version > zxdg_output_manager_v1_interface.version) {
@@ -231,7 +263,6 @@ struct wlr_xdg_output_manager_v1 *wlr_xdg_output_manager_v1_create(
 		return NULL;
 	}
 
-	wl_list_init(&manager->resources);
 	wl_list_init(&manager->outputs);
 	struct wlr_output_layout_output *layout_output;
 	wl_list_for_each(layout_output, &layout->outputs, link) {
@@ -246,21 +277,9 @@ struct wlr_xdg_output_manager_v1 *wlr_xdg_output_manager_v1_create(
 	wl_signal_add(&layout->events.change, &manager->layout_change);
 	manager->layout_destroy.notify = handle_layout_destroy;
 	wl_signal_add(&layout->events.destroy, &manager->layout_destroy);
+
+	manager->display_destroy.notify = handle_display_destroy;
+	wl_display_add_destroy_listener(display, &manager->display_destroy);
+
 	return manager;
 }
-
-void wlr_xdg_output_manager_v1_destroy(struct wlr_xdg_output_manager_v1 *manager) {
-	struct wlr_xdg_output_v1 *output, *output_tmp;
-	wl_list_for_each_safe(output, output_tmp, &manager->outputs, link) {
-		output_destroy(output);
-	}
-	struct wl_resource *resource, *resource_tmp;
-	wl_resource_for_each_safe(resource, resource_tmp, &manager->resources) {
-		wl_resource_destroy(resource);
-	}
-	wlr_signal_emit_safe(&manager->events.destroy, manager);
-	wl_list_remove(&manager->layout_add.link);
-	wl_list_remove(&manager->layout_change.link);
-	wl_list_remove(&manager->layout_destroy.link);
-	free(manager);
-}
